/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>

#include "mandoc.h"
#include "extern.h"

static	enum mlevel 	mandoc_level(struct mparse *mp, enum merror e);

static	const char * const errors[MERROR__MAX] = {
	"debug", /* MERROR_OK */
	/* MERROR_WARNING */
	"unterminated quoted string", /* MERROR_QUOTE */
	"ignoring extra arguments", /* MERROR_ARGSMANY */
	"skipping unknown escape", /* MERROR_ESCAPE */
	"end of line escape at end of file", /* MERROR_EEOF */
	"using gzip'd filename for `so'", /* MERROR_USEGZIP */
	"text before mdoc(7) or man(7) context", /* MERROR_TXT_NO_CTX */
	"macro scope broken", /* MERROR_BROKEN_SCOPE */
	"end of unopened scope", /* MERROR_BAD_SCOPE */
	/* MERROR_ERROR */
	"skipping unknown macro", /* MERROR_MACRO */
	"bad escape sequence", /* MERROR_BADESCAPE */
	"bad input encoding", /* MERROR_WCS */
	"unterminated multiline macro", /* MERROR_EXITSCOPE */
	"too few arguments", /* MERROR_ARGSFEW */
	"truncating malformed macro name", /* MERROR_NAME */
	"truncating malformed macro definition", /* MERROR_NAMEDEF */
	"ignoring malformed string definition", /* MERROR_STRINGDEF */
	"maximum stack depth exceeded", /* MERROR_STACK */
	"man(7) macro in mdoc(7) context", /* MERROR_MAN_IN_MDOC */
	"mdoc(7) macro in man(7) context", /* MERROR_MDOC_IN_MAN */
	/* MERROR_FATAL */
	"open", /* MERROR_OPEN */
	"stat", /* MERROR_STAT */
	"zlib", /* MERROR_ZLIB */
	"mmap", /* MERROR_MMAP */
	"read", /* MERROR_READ */
};

/*
 * Given a warning or error, adjust our highest level of message.
 * Return whether we should continue printing the message.
 */
static enum mlevel
mandoc_level(struct mparse *mp, enum merror e)
{

	/* Remember to order these highest--lowest. */

	if (e >= MERROR_FATAL) {
		if (mp->level < MLEVEL_FATAL)
			mp->level = MLEVEL_FATAL;
		return(MLEVEL_FATAL);
	} else if (e >= MERROR_ERROR) {
		if (mp->level < MLEVEL_ERROR)
			mp->level = MLEVEL_ERROR;
		return(MLEVEL_ERROR);
	} else if (e >= MERROR_WARNING) {
		if (mp->level < MLEVEL_WARNING)
			mp->level = MLEVEL_WARNING;
		return(MLEVEL_WARNING);
	}

	return(MLEVEL_OK);
}

void
merr(struct mparse *mp, enum merror e, 
		const char *f, size_t ln, size_t col)
{
	int		 err = errno;
	wchar_t		 buf[1024];
	enum mlevel	 l;

	if ((l = mandoc_level(mp, e)) < mp->wlevel)
		return;

	if (ln > 0)
		swprintf(buf, 1024, L"%s:%zu:%zu: %s: %s", 
			f, ln, col, errors[e], strerror(err));
	else
		swprintf(buf, 1024, L"%s: %s: %s",
			f, errors[e], strerror(err));

	(*mp->msg)(l, buf);
}

static void
merrxv(struct mparse *mp, enum merror e, enum mlevel l, const char *f, 
		size_t ln, size_t col, const wchar_t *fmt, va_list ap)
{
	wchar_t		 buf[1024];
	size_t		 sz;

	if (ln > 0)
		swprintf(buf, 1024, L"%s:%zu:%zu: %s", 
			f, ln, col, errors[e]);
	else
		swprintf(buf, 1024, L"%s: %s", f, errors[e]);

	if (NULL != fmt) {
		sz = wcslcat(buf, L": ", 1024);
		vswprintf(buf + sz, 1024 - sz, fmt, ap);
	}

	(*mp->msg)(l, buf);
}

void
merrx(struct mparse *mp, enum merror e, const char *f, 
		size_t ln, size_t col, const wchar_t *fmt, ...)
{
	va_list		 ap;
	enum mlevel	 lvl;

	if ((lvl = mandoc_level(mp, e)) >= mp->wlevel) {
		va_start(ap, fmt);
		merrxv(mp, e, lvl, f, ln, col, fmt, ap);
		va_end(ap);
	}
}

void
mlerrx(struct mline *l, enum merror e, 
		size_t col, const wchar_t *fmt, ...)
{
	va_list		 ap;
	enum mlevel	 lvl;
	struct mparse	*mp = l->file->parse;

	if ((lvl = mandoc_level(mp, e)) >= mp->wlevel) {
		va_start(ap, fmt);
		merrxv(mp, e, lvl, l->file->fname, 
			l->lineno, col, fmt, ap);
		va_end(ap);
	}
}
