/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>
#include <stdlib.h>
#include <wchar.h>

#include "mandoc.h"
#include "extern.h"

struct	mmacrodef {
	enum mdtype	 type; /* macro type */
	mmacro		 macro; /* processing function */
	mlparse		 parser; /* argument parsing function */
	const wchar_t	*name; /* name (e.g., `cc') */
};

static	void	*halloc(size_t n, void *arg);
static	void	 hfree(void *p, size_t n, void *arg);

static	const struct mmacrodef macrodefs[] = {
	{ MDTYPE_Ac, mmacro_mdoc, NULL, L"Ac" },
	{ MDTYPE_Ao, mmacro_mdoc, NULL, L"Ao" },
	{ MDTYPE_Aq, mmacro_mdoc, NULL, L"Aq" },
	{ MDTYPE_Fl, mmacro_mdoc, NULL, L"Fl" },
	{ MDTYPE_Sh, mmacro_mdoc, NULL, L"Sh" },
	{ MDTYPE_Ss, mmacro_mdoc, NULL, L"Ss" },
	{ MDTYPE_TH, mmacro_man, NULL, L"TH" },
	{ MDTYPE_Xc, mmacro_mdoc, NULL, L"Xc" },
	{ MDTYPE_Xo, mmacro_mdoc, NULL, L"Xo" },
	{ MDTYPE_c2, mmacro_char, mline_cc, L"c2" },
	{ MDTYPE_cc, mmacro_char, mline_cc, L"cc" },
	{ MDTYPE_de, mmacro_de, mline_de, L"de" },
	{ MDTYPE_de1, mmacro_de, mline_de, L"de1" },
	{ MDTYPE_ds, mmacro_ds, mline_ds, L"ds" },
	{ MDTYPE_ds1, mmacro_ds, mline_ds, L"ds1" },
	{ MDTYPE_ec, mmacro_char, mline_cc, L"ec" },
	{ MDTYPE_ecr, mmacro_ecr, mline_ecr, L"ecr" },
	{ MDTYPE_ecs, mmacro_char, mline_cc, L"ecs" },
	{ MDTYPE_el, mmacro_el, mline_el, L"el" },
	{ MDTYPE_ie, mmacro_ie, mline_if, L"ie" }, 
	{ MDTYPE_if, mmacro_if, mline_if, L"if" }, 
	{ MDTYPE_ig, mmacro_ig, mline_de, L"ig" }, 
	{ MDTYPE_Oc, mmacro_mdoc, NULL, L"Oc" },
	{ MDTYPE_Oo, mmacro_mdoc, NULL, L"Oo" },
	{ MDTYPE_Op, mmacro_mdoc, NULL, L"Op" },
	{ MDTYPE_so, mmacro_so, NULL, L"so" },
	{ MDTYPE_while, mmacro_while, mline_if, L"while" },
	{ 0, NULL, NULL, NULL },
};

static void
hfree(void *p, size_t n, void *arg)
{

	free(p);
}

static void *
halloc(size_t n, void *arg)
{

	return(mandoc_calloc(n, 1));
}

unsigned int
mdmacro_lookup(struct mparse *mp, const wchar_t *name)
{
	const char	*start, *end;

	start = (char *)name;
	end = start + (wcslen(name) * sizeof(wchar_t));
	return(ohash_qlookupi(&mp->macros, start, &end));
}

unsigned int
mdmacro_lookups(struct mparse *mp, 
		const wchar_t *start, const wchar_t *end)
{

	return(ohash_qlookupi(&mp->macros, 
		(char *)start, (const char **)&end));
}

struct mdmacro *
mdmacro_find(struct mparse *mp, const wchar_t *name)
{

	return(ohash_find(&mp->macros, mdmacro_lookup(mp, name)));
}

struct mdmacro *
mdmacro_finds(struct mparse *mp, 
		const wchar_t *start, const wchar_t *end)
{

	return(ohash_find(&mp->macros, mdmacro_lookups(mp, start, end)));
}

struct mdmacro *
mdmacro_calloc(const wchar_t *name)
{
	size_t		 wsz, sz;
	struct mdmacro	*m;

	wsz = wcslen(name);
	sz = (wsz + 1) * sizeof(wchar_t);
	m = mandoc_calloc(1, sizeof(struct mdmacro) + sz);
	wcslcpy(m->name, name, wsz + 1);
	return(m);
}

void
mdmacro_init(struct mparse *mp)
{
	struct ohash_info info;
	struct mdmacro	 *md;
	size_t		  i;
	unsigned int	  idx;

	info.halloc = info.alloc = halloc;
	info.hfree = hfree;
	info.key_offset = offsetof(struct mdmacro, name);

	ohash_init(&mp->macros, 4, &info);

	/*
	 * Add the default set of macros (i.e., the mdoc, man, and
	 * whichever other macros) to the set of parsed ones.
	 */
	for (i = 0; NULL != macrodefs[i].name; i++) {
		md = mdmacro_calloc(macrodefs[i].name);
		md->type = macrodefs[i].type;
		md->macro = macrodefs[i].macro;
		md->parser = macrodefs[i].parser;
		idx = mdmacro_lookup(mp, md->name);
		assert(NULL == ohash_find(&mp->macros, idx));
		ohash_insert(&mp->macros, idx, md);
	}
}

void
mdmacro_clear(struct mparse *mp)
{
	struct mdmacro	*md;
	unsigned int	 idx;

	/* 
	 * If we call mdmacro_clear() twice in a row, this would segfault
	 * because ohash_delete() NULLifies the table but not the table size,
	 * and ohash_next() uses the table size.
	 * Protect against this with an assertion.
	 */
	assert(NULL != mp->macros.t);

	for (md = ohash_first(&mp->macros, &idx); NULL != md; 
			md = ohash_next(&mp->macros, &idx)) 
		free(md);

	ohash_delete(&mp->macros);
}
