/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/param.h>

#include <assert.h>
#include <stdlib.h>
#include <wchar.h>

#include "mandoc.h"
#include "extern.h"

union	mend {
	enum mdtype	 macro;
	size_t		 count;
};

/*
 * A mdoc(7) macro handler.
 * Returns 1 if we should adjust "lp" to the next line, or 0 if *lp is
 * set to the next line already.
 * "type" is set to the parent's invocation type.
 */
typedef	int (*mmdoc)(struct mline **lp, enum mdtype type);

enum	mendtype {
	MEND_DEFAULT,
	MEND_MACRO,
	MEND_COUNT
};

struct	mdoc {
	enum mendtype	 endtype;
	union mend	 end;
	unsigned int	 fl;
#define	MDOC_CALL	 1 /* can be called within parsed macro */
#define	MDOC_PARSE	 2 /* arguments parsed for more macros */
#define MDOC_WEAK	 4 /* multiple line with weak scope */
#define	MDOC_WEAK_END	 8 /* explicit end of weak block */
#define	MDOC_BLOCK	 16 /* line-blocked */
#define MDOC_SPEC	 32 /* specially-handled node */
#define	MDOC_NORM	 MDOC_PARSE|MDOC_CALL
#define	MDOC_PBLOCK	 MDOC_BLOCK|MDOC_PARSE|MDOC_CALL
#define	MDOC_WCLOSE	 MDOC_WEAK_END|MDOC_CALL|MDOC_PARSE
#define	MDOC_WOPEN	 MDOC_WEAK|MDOC_CALL|MDOC_PARSE
	mmdoc		 macro;
};

static	int	mdoc_callable(struct mline *l, 
			size_t *pos, enum mdtype *type);
static	int	mdoc_block(struct mline **lp, enum mdtype type);
static	int	mdoc_block_inline(struct mline **lp, 
			size_t *pos, enum mdtype type);
static	int	mdoc_line(struct mline **lp, enum mdtype type);
static	int	mdoc_line_inblock(struct mline **lp, 
			size_t *pos, enum mdtype type);
static	void	mdoc_next(struct mline **lp, enum mdtype type);
static	void	mdoc_nextline(struct mline **lp);
static	int	mdoc_Sh(struct mline **lp, enum mdtype type);
static	int	mdoc_Ss(struct mline **lp, enum mdtype type);
static	int	mdoc_weak(struct mline **lp, enum mdtype type);
static	int	mdoc_weak_inblock(struct mline **lp, 
			size_t *pos, enum mdtype type);
static	void	mdoc_weak_pop(struct mline *l, size_t pos,
			struct mnode *n, enum mdtype type);

static	const struct mdoc mdocs[MDTYPE__MAX] = {
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__TEXT */
	{ MEND_COUNT, { 0 }, MDOC_WCLOSE, mdoc_line }, /* MDTYPE_Ac */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ad */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_An */
	{ MEND_MACRO, { MDTYPE_Ac }, MDOC_WOPEN, mdoc_weak }, /* MDTYPE_Ao */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ap */
	{ MEND_MACRO, { MDTYPE_Ac }, MDOC_PBLOCK, mdoc_block }, /* MDTYPE_Aq */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ar */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_At */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bc */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bd */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bf */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bk */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bl */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bo */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bq */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Brc */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bro */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Brq */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bsx */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bt */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Bx */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Cd */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Cm */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_D1 */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Db */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Dc */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Dd */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Dl */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Do */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Dq */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Dt */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Dv */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Dx */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ec */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ed */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ef */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ek */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_El */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Em */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_En */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Eo */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Er */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Es */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ev */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ex */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Fa */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Fc */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Fd */
	{ MEND_DEFAULT, { MDTYPE__MAX }, MDOC_NORM, mdoc_line }, /* MDTYPE_Fl */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Fn */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Fo */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Fr */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ft */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Fx */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Hf */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ic */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_In */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_It */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Lb */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Li */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Lk */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Lp */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ms */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Mt */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Nd */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Nm */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_No */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ns */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Nx */
	{ MEND_COUNT, { 0 }, MDOC_WCLOSE, mdoc_line }, /* MDTYPE_Oc */
	{ MEND_MACRO, { MDTYPE_Oc }, MDOC_WOPEN, mdoc_weak }, /* MDTYPE_Oo */
	{ MEND_DEFAULT, { MDTYPE__MAX }, MDOC_PBLOCK, mdoc_block }, /* MDTYPE_Op */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Os */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ot */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ox */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Pa */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Pc */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Pf */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Po */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Pp */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Pq */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Qc */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ql */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Qo */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Qq */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Re */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Rs */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Rv */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Sc */
	{ MEND_DEFAULT, { MDTYPE__MAX }, MDOC_SPEC, mdoc_Sh }, /* MDTYPE_Sh */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Sm */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_So */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Sq */
	{ MEND_DEFAULT, { MDTYPE__MAX }, MDOC_SPEC, mdoc_Ss }, /* MDTYPE_Ss */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_St */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Sx */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Sy */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_TH */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ta */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Tn */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ud */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Ux */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Va */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Vt */
	{ MEND_COUNT, { 0 }, MDOC_WCLOSE, mdoc_line }, /* MDTYPE_Xc */
	{ MEND_MACRO, { MDTYPE_Ac }, MDOC_WOPEN, mdoc_weak }, /* MDTYPE_Xo */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_Xr */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__A */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__B */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__C */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__D */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__I */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__J */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__N */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__O */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__P */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__Q */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__R */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__T */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__U */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__V */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_br */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_c2 */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_cc */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_de */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_de1 */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_ds */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_ds1 */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_ec */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_ecr */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_ecs */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_el */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_ie */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_if */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_ig */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_so */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_sp */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE_while */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__USERDEF */
	{ MEND_DEFAULT, { MDTYPE__MAX }, 0, NULL }, /* MDTYPE__NONE */
};

void
mmacro_mdoc(struct mline **lp)
{
	struct mparse	*mp = (*lp)->file->parse;
	struct mline	*l = *lp;
	struct mnode	*n;

	if (MINPUT_MAN == mp->input) {
		mlerrx(l, MERROR_MAN_IN_MDOC,
			l->col[0], L"%ls", l->argv[0]);
		mdoc_nextline(lp);
		return;
	} else if (MINPUT_UNSET == mp->input)
		mp->input = MINPUT_MDOC;

	if (NULL != (n = mp->node)) 
		assert(MELEM__ROOT == n->elem);

	assert(NULL != mdocs[l->macro].macro);
	if (mdocs[l->macro].macro(lp, MDTYPE__NONE))
		mdoc_nextline(lp);
	mp->node = NULL == n ? mp->root : n;
}

static void
mdoc_next(struct mline **lp, enum mdtype type)
{
	struct mnode	*save = (*lp)->file->parse->node;

	if (NULL != mdocs[(*lp)->macro].macro) {
		if (mdocs[(*lp)->macro].macro(lp, type))
			mdoc_nextline(lp);
	} else
		mmacro_next(lp);

	if (NULL != *lp)
		(*lp)->file->parse->node = save;
}

static void
mdoc_nextline(struct mline **lp)
{
	size_t		 i;
	struct mparse	*mp;
	struct mnode	*top;

	if (NULL == *lp) 
		return;
	mp = (*lp)->file->parse;
	top = mp->node;
	for (i = 0; i < mp->lstackpos; i++) {
		if (MDTYPE__MAX == mp->lstack[i].type) 
			continue;
		mlerrx(mp->lstack[i].line, 
			MERROR_BROKEN_SCOPE, 
			mp->lstack[i].n->pos, NULL);
		mnode_alloc_node
			(mp->lstack[i].line, 0,
			 mdocs[mp->lstack[i].type].end.macro);
		mp->lstack[i].n->broken = mp->node;
		mp->lstack[i].type = MDTYPE__MAX;
		mp->node = top;
	}
	*lp = mline_popnext(*lp);
}

static int
mdoc_Sh(struct mline **lp, enum mdtype type)
{

	mnode_alloc_node_line(*lp, MDTYPE_Sh);
	mnode_alloc_text_all(*lp);
	mdoc_nextline(lp);

	while (NULL != *lp) {
		mline_parse(*lp);
		if (MDTYPE_Sh == (*lp)->macro) 
			break;
		mdoc_next(lp, MDTYPE_Sh);
	}

	return(0);
}

static int
mdoc_Ss(struct mline **lp, enum mdtype type)
{

	mnode_alloc_node_line(*lp, MDTYPE_Ss);
	mnode_alloc_text_all(*lp);
	mdoc_nextline(lp);

	while (NULL != *lp) {
		mline_parse(*lp);
		if (MDTYPE_Sh == (*lp)->macro)
			break;
		else if (MDTYPE_Ss == (*lp)->macro)
			break;
		mdoc_next(lp, MDTYPE_Ss);
	}

	return(0);
}

static int
mdoc_line(struct mline **lp, enum mdtype type)
{
	size_t		 pos = 0;

	if ( ! mdoc_line_inblock(lp, &pos, type)) {
		assert(0 == pos);
		return(0);
	} else if (NULL != *lp)
		assert(pos == (*lp)->argc);

	return(1);
}

/*
 * Handle a line beginning with an inline macro OR a non-callable
 * argument within a parsed line.
 * For example, `.Fl foo' would be handled here, as would the trailing
 * part of `.Ao foo Ac bar' (which might continue with another multiline
 * block scope like with Oo).
 */
static int
mdoc_line_inblock(struct mline **lp, size_t *pos, enum mdtype type)
{
	struct mparse	*mp = (*lp)->file->parse;
	struct mnode	*save;
	enum mdtype	 savetype = type;
	size_t		 mpos;
	int		 rc = 1;

	save = mp->node;
	mpos = mp->wstackpos;

	if (0 == *pos) {
		mnode_alloc_node(*lp, (*pos)++, (*lp)->macro);
		type = (*lp)->macro;
		if (MDOC_WEAK_END & mdocs[type].fl)
			mdoc_weak_pop(*lp, *pos - 1, mp->node, type);
	}

	while (rc > 0 && NULL != *lp && *pos < (*lp)->argc) {
		assert(mp->wstackpos == mpos);
		if ( ! mdoc_callable(*lp, pos, &type))
			continue;
		mp->node = save;
		if (MDTYPE__TEXT == type) {
			type = savetype;
			continue;
		} else if (MDOC_WEAK & mdocs[type].fl) {
			rc = mdoc_weak_inblock(lp, pos, type);
			continue;
		} else if (MDOC_BLOCK & mdocs[type].fl) {
			rc = mdoc_block_inline(lp, pos, type);
			continue;
		}
		mnode_alloc_node(*lp, (*pos)++, type);
		if (MDOC_WEAK_END & mdocs[type].fl)
			mdoc_weak_pop(*lp, *pos - 1, mp->node, type);
	}

	assert(mp->wstackpos == mpos);
	return(rc);
}

static int
mdoc_weak(struct mline **lp, enum mdtype type)
{
	size_t		 pos = 0;

	if ( ! mdoc_weak_inblock(lp, &pos, (*lp)->macro)) {
		assert(0 == pos);
		return(0);
	} else if (NULL != *lp) {
		mdoc_line_inblock(lp, &pos, type);
		assert(pos == (*lp)->argc);
	}

	return(1);
}

/*
 * Handle a weakly-scoped macro like Ao that can have multiple lines.
 * After opening the scope, we walk through lines, parsing them, til we
 * get to one that has our valid closure.
 * Since our closure can occur in a nested context, we monitor our
 * current stack entry to see if it's been zeroed.
 * This function is also a bit complicated because we can be closed out
 * by scopes such as Ss.
 */
static int
mdoc_weak_inblock(struct mline **lp, size_t *pos, enum mdtype type)
{
	struct mparse	*mp = (*lp)->file->parse;
	struct mnode	*save, *top = (*lp)->file->parse->node;
	enum mdtype	 close, savetype = type;
	size_t		 wpos;
	int		 rc = 1;

	assert(MEND_MACRO == mdocs[type].endtype);
	close = mdocs[type].end.macro;
	mnode_alloc_node(*lp, (*pos)++, type);
	wpos = mp->wstackpos++;
	mp->wstack[wpos].type = type;
	mp->wstack[wpos].line = *lp;
	mp->wstack[wpos++].n = save = mp->node;

	while (rc > 0 && NULL != *lp) {
		while (NULL != *lp && rc > 0 && *pos < (*lp)->argc) {
			assert(mp->wstackpos == wpos);
			if (MDTYPE__MAX == mp->wstack[wpos - 1].type)
				break;
			if ( ! mdoc_callable(*lp, pos, &type))
				continue;
			mp->node = save;
			if (MDTYPE__TEXT == type) {
				type = savetype;
				continue;
			} else if (MDOC_WEAK & mdocs[type].fl) {
				rc = mdoc_weak_inblock(lp, pos, type);
				continue;
			} else if (MDOC_BLOCK & mdocs[type].fl) {
				rc = mdoc_block_inline(lp, pos, type);
				continue;
			}
			mnode_alloc_node(*lp, (*pos)++, type);
			if (close == type) {
				mdoc_callable(*lp, pos, &type);
				mp->wstack[mp->wstackpos - 1].type = MDTYPE__MAX;
				break;
			} else if (MDOC_WEAK_END & mdocs[type].fl)
				mdoc_weak_pop(*lp, *pos - 1, mp->node, type);
		}
		if (0 == rc || NULL == *lp || *pos < (*lp)->argc)
			break;
		if (MDTYPE__MAX == mp->wstack[wpos - 1].type)
			break;
		/*
		 * Handle the fact that subsequent lines need not be
		 * mdoc(7) lines: they can be text or other macros.
		 */
		*pos = 0;
		mdoc_nextline(lp);
		while (NULL != *lp) {
			mline_parse(*lp);
			if (NULL == mdocs[(*lp)->macro].macro) {
				mdoc_next(lp, savetype);
				continue;
			} else if (MDOC_SPEC & mdocs[(*lp)->macro].fl) 
				rc = 0;
			break;
		}
	}

	mp->wstackpos--;
	assert(mp->wstackpos == --wpos);
	if (MDTYPE__MAX != mp->wstack[wpos].type) {
		mnode_alloc_node
			(mp->wstack[wpos].line, 0, close);
		mp->wstack[wpos].n->broken = mp->node;
		mp->wstack[wpos].type = MDTYPE__MAX;
	}
	mp->node = top;
	return(rc);
}

static int
mdoc_block(struct mline **lp, enum mdtype type)
{
	size_t		 pos = 0;

	if ( ! mdoc_block_inline(lp, &pos, (*lp)->macro)) {
		assert(0 == pos);
		return(0);
	} else if (NULL != *lp) {
		mdoc_line_inblock(lp, &pos, type);
		assert(pos == (*lp)->argc);
	}

	return(1);
}

/*
 * Process a line-scoped block like Aq.
 * This extends to the end of the current line, so we set a line trap
 * (lstack) with the macro.
 * However, the real scope can be extended with recursive weakly-scoped
 * macros (try putting an Ao within an Aq to see this effect).
 */
static int
mdoc_block_inline(struct mline **lp, size_t *pos, enum mdtype type)
{
	struct mparse	*mp = (*lp)->file->parse;
	struct mline	*l = *lp;
	struct mnode	*top, *save;
	enum mdtype	 savetype = type;
	size_t		 wpos = (*lp)->file->parse->wstackpos, 
			 lpos = (*lp)->file->parse->lstackpos;
	int		 rc = 1;

	top = mp->node;
	mnode_alloc_node(*lp, (*pos)++, type);
	mp->lstack[mp->lstackpos].n = save = mp->node;
	mp->lstack[mp->lstackpos].line = *lp;
	mp->lstack[mp->lstackpos++].type = type;

	while (rc > 0 && l == *lp && *pos < (*lp)->argc) {
		assert(mp->wstackpos == wpos);
		if ( ! mdoc_callable(*lp, pos, &type)) 
			continue;
		mp->node = save;
		if (MDTYPE__TEXT == type) {
			type = savetype;
			continue;
		} else if (MDOC_WEAK & mdocs[type].fl) {
			rc = mdoc_weak_inblock(lp, pos, type);
			continue;
		} else if (MDOC_BLOCK & mdocs[type].fl) {
			rc = mdoc_block_inline(lp, pos, type);
			continue;
		}
		mnode_alloc_node(*lp, (*pos)++, type);
		if (MDOC_WEAK_END & mdocs[type].fl)
			mdoc_weak_pop(*lp, *pos - 1, mp->node, type);
	}

	mp->lstackpos--;
	assert(mp->lstackpos == lpos);
	assert(mp->wstackpos == wpos);
	if (MDTYPE__MAX != mp->lstack[mp->lstackpos].type) {
		assert(savetype == mp->lstack[mp->lstackpos].type);
		mnode_alloc_node
			(mp->lstack[mp->lstackpos].line, 
			 0, mdocs[savetype].end.macro);
		mp->lstack[mp->lstackpos].n->broken = mp->node;
		mp->lstack[mp->lstackpos].type = MDTYPE__MAX;
	}
	
	mp->node = top;
	return(rc);
}

/*
 * Process text node until a hard limit to the current macro's arguments
 * is reached, we reach another callable macro, or the end of line.
 * Returns 1 if encounters another macro.
 * Returns 0 if at eoln.
 * Sets "type" to be the new macro's type or MDTYPE__TEXT if instead we
 * exhaust the caller's type's macro count (e.g., the Ac macro).
 */
static int
mdoc_callable(struct mline *l, size_t *pos, enum mdtype *type)
{
	struct mdmacro	*md;
	struct mparse	*mp = l->file->parse;
	size_t		 lim;

	lim = MEND_COUNT == mdocs[*type].endtype ? 
		*pos + mdocs[*type].end.count : l->argc;

	if (lim > l->argc)
		lim = l->argc;

	while  (*pos < lim) {
		md = mdmacro_find(mp, l->argv[*pos]);
		if (NULL != md && MDOC_CALL & mdocs[md->type].fl) {
			*type = md->type;
			return(1);
		}
		mnode_alloc_text(l, (*pos)++);
	}

	*type = MDTYPE__TEXT;
	return(*pos < l->argc);
}

/*
 * Call this when unexpectedly encountering a closing macro.
 * "Expected" means it's symmetrically under a weak opening macro.
 * If we find the scope, we mark it as prematurely closed.
 * If we don't, we emit a warning.
 */
static void
mdoc_weak_pop(struct mline *l, size_t pos,
		struct mnode *n, enum mdtype type)
{
	size_t		 i;
	struct mstacked	*ms = l->file->parse->wstack;
	
	for (i = l->file->parse->wstackpos; i > 0; i--) {
		assert(MEND_MACRO == mdocs[ms[i - 1].type].endtype);
		if (mdocs[ms[i - 1].type].end.macro != type)
			continue;
		mlerrx(ms[i - 1].line, 
			MERROR_BROKEN_SCOPE, 
			ms[i - 1].n->pos, NULL);
		ms[i - 1].type = MDTYPE__MAX;
		ms[i - 1].n->broken = n;
		return;
	} 
	mlerrx(l, MERROR_BAD_SCOPE, l->col[pos], NULL);
}
