/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/param.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <wchar.h>

#include <zlib.h>

#include "mandoc.h"
#include "extern.h"

#define	MFILE_ZGROWSZ	 2 /* grow multiplier of zlib space */

static	struct mfile	*mfile_new(struct mparse *mp, 
				int *err, const char *file);

/*
 * Free a file and all of its lines, parsed argument, and so on.
 * It also frees the "mf" pointer itself, which can be NULL.
 */
void
mfile_free(struct mfile *mf)
{
	size_t		 i;

	if (NULL == mf)
		return;

	for (i = 0; i < mf->sz; i++)
		mline_free(&mf->lines[i]);

	free(mf->lines);
	free(mf->fbuf);
	free(mf->fname);
	free(mf);
}

/*
 * Parse a (possibly gzip'd) file into an ordered sequence of lines.
 * These are managed by the allocated mfile.
 * In reality, the mfile lines' wchar_t data all point into a single
 * buffer of the newline-delimited data ("fbuf").
 * Both \n and \r\n formats are handled.
 */
static struct mfile *
mfile_new(struct mparse *mp, int *err, const char *file)
{
	const char	*suffix;
	int		 fd = -1, zc = 0;
	struct mfile	*mf = NULL;
	struct stat	 st;
	char		*mb, *buf = NULL, *zbuf = NULL;
	wchar_t		*cp, *start;
	z_stream	 zstr;
	size_t	 	 sz = 0, zsz, fsz, lines, len;

	/* Support gzip'd files by detecting the standard .gz. */

	if (NULL != (suffix = strrchr(file, '.')) &&
			 0 == strcasecmp(suffix + 1, "gz"))
		zc = 1;

	if (-1 == (fd = open(file, O_RDONLY, 0))) {
		/*
		 * Don't report the error: the caller does this.
		 * Save our errno for processing.
		 */
		*err = errno;
		goto err;
	} else if (-1 == fstat(fd, &st)) {
		merr(mp, MERROR_STAT, file, 0, 0);
		goto err;
	} else if ( ! S_ISREG(st.st_mode)) {
		merrx(mp, MERROR_READ, file, 0, 0, L"not regular");
		goto err;
	} else if (st.st_size >= (1U << 31)) {
		merrx(mp, MERROR_READ, file, 0, 0, L"file too long");
		goto err;
	} else if (0 == st.st_size) {
		merrx(mp, MERROR_READ, file, 0, 0, L"empty file");
		goto err;
	}

	/* Now mmap() the entire file into read-only memory. */

	zsz = sz = (size_t)st.st_size;
	zbuf = buf = mmap(NULL, sz, 
		PROT_READ, MAP_FILE|MAP_SHARED, fd, 0);

	if (buf == MAP_FAILED) {
		merr(mp, MERROR_MMAP, file, 0, 0);
		buf = zbuf = NULL;
		goto err;
	}

	mf = mandoc_calloc(1, sizeof(struct mfile));
	mf->fname = mandoc_strdup(file);
	mf->parse = mp;

	/* Use zlib to decompress the gzip file? */

	if (zc) {
		zbuf = NULL;
		/*
		 * Double the inflated buffer size each time we don't
		 * have enough.
		 */
		zsz = sz;
		do {
			memset(&zstr, 0, sizeof(zstr));
			zstr.next_in = (Bytef *)buf;
			zstr.avail_in = sz;
			if (Z_OK != inflateInit2(&zstr, 32 + MAX_WBITS)) {
				merrx(mp, MERROR_ZLIB, file, 0, 0, NULL);
				goto err;
			}
			zsz *= 2;
			zbuf = mandoc_realloc(zbuf, zsz + 1);
			zstr.next_out = (Bytef *)zbuf;
			zstr.avail_out = zsz;
			zc = inflate(&zstr, Z_FINISH);
			inflateEnd(&zstr);
		} while (Z_BUF_ERROR == zc);

		if (Z_STREAM_END != zc) {
			merrx(mp, MERROR_ZLIB, file, 0, 0, NULL);
			goto err;
		}

		zsz = zstr.total_out;
	} 

	/*
	 * Copy the mapped/gunzip'd memory image into the appropriate
	 * wide-character glyphs.
	 * Then close out hte mmap and free gunzip'd bytes.
	 * We use mblen() instead of mbstowcs() because the string buf
	 * isn't guaranteed to be nil-terminated.
	 */
	for (fsz = 0, mb = zbuf; zsz > 0; fsz++) {
		if (-1 == (zc = mblen(mb, zsz))) {
			merrx(mp, MERROR_WCS, mf->fname, 0, 0, NULL);
			goto err;
		}
		assert(zc > 0);
		mb += zc;
		zsz -= zc;
	}

	mf->fbuf = mandoc_malloc((fsz + 1) * sizeof(wchar_t));
	mbstowcs(mf->fbuf, zbuf, fsz);
	mf->fbuf[fsz] = L'\0';

	if (zbuf != buf)
		free(zbuf);

	munmap(buf, sz);
	close(fd);

	/*
	 * Count the number of newlines.
	 * We'll use this to allocate our line buffers.
	 */
	for (cp = mf->fbuf; NULL != (cp = wcschr(cp, L'\n')); cp++)
		mf->sz++;
	mf->lines = mandoc_calloc(mf->sz, sizeof(struct mline));

	/*
	 * Assign each line to a nil-terminated line in the input.
	 * Set our initial line ordering and positions, too.
	 * Strip away \r if we encounter them.
	 */
	start = mf->fbuf;
	lines = 0;
	while (NULL != (cp = wcschr(start, L'\n'))) {
		mf->lines[lines].next[0] = &mf->lines[lines + 1];
		mf->lines[lines].line = start;
		mf->lines[lines].lineno = lines + 1;
		mf->lines[lines].file = mf;
		/* avoid wcslen() */
		len = cp - start; 
		if (cp > start && L'\r' == cp[-1]) {
			len--;
			cp[-1] = L'\0';
		}
		*cp++ = L'\0';
		/*mf->lines[lines].len = len;*/
		start = cp;
		lines++;
	}
	if (lines > 0)
		mf->lines[lines-1].next[0] = NULL;

	return(mf);
err:
	if (zbuf != buf)
		free(zbuf);
	if (NULL != buf)
		munmap(buf, sz);
	if (-1 != fd)
		close(fd);
	mfile_free(mf);
	return(NULL);
}

/*
 * Parse the file "f" and adjust its last line to point to our next
 * line, then our next line to point to its first line.
 */
int
mfile_next(struct mparse *mp, const char *f, struct mline *l)
{
	struct mfile	*mf;
	char		 buf[MAXPATHLEN];
	int		 err;

	err = 0;
	if (NULL == (mf = mfile_new(mp, &err, f))) {
		if (ENOENT != err || NULL == l ||
				NULL != strstr(f, ".gz")) {
			merr(mp, MERROR_OPEN, f, 0, 0);
			return(0);
		}

		/*
		 * If we fail to open the file, try opening a gzip'd
		 * version of it.
		 * This only happens when parsing `so': groff's `so' has
		 * this behaviour, so we should too.
		 * But print an error message, anyway.
		 */
		assert(NULL != l);
		mlerrx(l, MERROR_USEGZIP, 0, L"%s", f);
		strlcpy(buf, f, MAXPATHLEN);
		strlcat(buf, ".gz", MAXPATHLEN);
		if (NULL == (mf = mfile_new(mp, &err, buf))) {
			merr(mp, MERROR_OPEN, f, 0, 0);
			return(0);
		}
	}

	mf->next = mp->file;
	mp->file = mf;

	/*
	 * If we're already parsing, set the first line of the
	 * newly-parsed file after the current line, and the last line
	 * of the newly-parsed file as the next line.
	 */
	if (NULL != l) {
		assert(l->file->parse == mp);
		mf->lines[mf->sz - 1].next[0] = l->next[0];
		l->next[0] = &mf->lines[0];
	} 

	return(1);
}

