/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>

#include "mandoc.h"
#include "extern.h"

#define	MLINE_GROWSZ	  32 /* grow adder of tokens */

static	void	 mline_growarg(struct mline *l, 
			const wchar_t *arg, size_t col);
static	void	 mline_strdup(struct mline *l);

/*
 * Grow the buffer of argvs and cols by a slop (MLINE_GROWSZ).
 * Set the last position to the given argument and column.
 */
static void
mline_growarg(struct mline *l, const wchar_t *arg, size_t col)
{

	if (l->argc + 1 > l->maxarg) {
		l->maxarg = l->argc + 1 + MLINE_GROWSZ;
		l->argv = mandoc_realloc
			(l->argv, l->maxarg * sizeof(wchar_t *));
		l->col = mandoc_realloc
			(l->col, l->maxarg * sizeof(size_t));
	}
	l->argv[l->argc] = arg;
	l->col[l->argc++] = col;
}

/*
 * Get the next line in a parse sequence.
 * This accounts for recursive calls, where the next line may differ
 * between recursive calls to the same macro.
 * NULL current lines return NULL next lines.
 */
struct mline *
mline_popnext(struct mline *l)
{

	assert(NULL != l);
	return(l->stackpos > 0 ? l->next[l->stackpos--] : l->next[0]);
}

struct mline *
mline_peek(struct mline *l)
{

	return(l->stackpos > 0 ? l->next[l->stackpos] : l->next[0]);
}

/*
 * Set the next line in a recursive parse sequence.
 * This increases the stack depth of the current line.
 * Return 0 if we've exceeded the stack depth.
 */
int
mline_pushnext(struct mline *l, struct mline *next)
{

	if (l->stackpos + 1 >= MLINE_STACKSZ) {
		mlerrx(l, MERROR_STACK, 1, NULL);
		return(0);
	} 

	l->next[++l->stackpos] = next;
	return(1);
}

/*
 * Free the contents of an mline.
 * We DO NOT free the line pointer itself.
 */
void
mline_free(struct mline *l)
{

	if (NULL != l) {
		free(l->copy);
		free(l->wline);
		free(l->argv);
		free(l->col);
	}
}

/*
 * Like wcsdup() but uses realloc() internally.
 * Use the copy-buffer if it exists (i.e., we've already been invoked in
 * a `de'), else use the real line.
 */
static void
mline_strdup(struct mline *l)
{
	size_t		 sz;
	const wchar_t	*cp;

	cp = NULL == l->copy ? l->line : l->copy;
	sz = (wcslen(cp) + 1) * sizeof(wchar_t);
	l->wline = mandoc_realloc(l->wline, sz);
	memcpy(l->wline, cp, sz);
}

/*
 * Take a nil-terminated macro line and parse it into tokens of argc and
 * argv, with column offsets of col.
 * Before tokenising, do all replacement of special strings.
 * The parse rules are derived from groff.texinfo.
 * Use the copy-buffer if it exists (i.e., we've already been invoked in
 * a `de'), else use the real line.
 */
void
mline_parse(struct mline *l)
{
	size_t		 i, start, rep;
	struct mdmacro	*md;
	wchar_t		*cp, *macro;
	const wchar_t	*line;
	struct mparse	*mp = l->file->parse;

	/*
	 * Use this very sparingly.
	 * It allows us to "peek" at the next macro's parsed content
	 * before invoking its handler.
	 * See mmacro_next() for usage.
	 */
	if (l->clean)
		return;

	/* Reset parsed arguments. */

	l->clean = 1;
	l->argc = 0;
	l->ctr = L'\0';
	l->macro = MDTYPE__TEXT;
	line = NULL == l->copy ? l->line : l->copy;

	/*
	 * We want to avoid copying around the string, so if we're
	 * neither a macro nor need replacements, use the original
	 * buffer directly for our arguments.
	 */
	if (mp->ctr != line[0] && 
			mp->lctr != line[0] &&
			NULL == wcschr(line, mp->esc)) {
		mline_growarg(l, line, 1);
		return;
	}

	/*
	 * We're either a macro or have glyphs.
	 * First unescape our line directly into wline.
	 * If we haven't unescaped anything, copy over.
	 */
	if (0 == (rep = munescape(l, &l->wline)))
		mline_strdup(l);

	/*
	 * Check again if we're a macro.
	 * If we're not, directly use the working-copy string for
	 * processing.
	 * XXX: this can be optimised: if we've made replacements but
	 * aren't a macro, there's no need to run mline_strdup() in the
	 * above conditional.
	 */
	if (mp->ctr != l->wline[0] && mp->lctr != l->wline[0]) {
		mline_growarg(l, l->wline, 1);
		return;
	} 

	cp = l->wline;
	l->ctr = l->wline[0];
	l->macro = MDTYPE__NONE;
	i = 1;

	/*
	 * According to groff.texinfo, there may be any number of spaces
	 * AND TABS between the control character and the macro.
	 */
	while (L' ' == cp[i] || L'\t' == cp[i])
		i++;

	/* Check: we're just a dot and whitespace. */

	if (L'\0' == cp[i])
		return;

	/*
	 * Macro names have non-standard syntax.
	 * For one, quote marks are considered opaque text.
	 * Second, escapes can't occur within the macro name.
	 * If they do, the macro name is truncated til the next
	 * space-separated word.
	 * Wait til we've looked up our macro before we add it to the
	 * argv array.
	 */
	start = i + 1;
	for (macro = &cp[i]; L'\0' != cp[i] && L' ' != cp[i]; i++)
		if (mp->esc == cp[i]) {
			mlerrx(l, MERROR_NAME, i + 1, NULL);
			cp[i++] = L'\0';
			while (L'\0' != cp[i] && L' ' != cp[i])
				i++;
			break;
		}

	if (L'\0' != cp[i])
		cp[i++] = L'\0';

	/* We just throw away empty or unfound macros. */

	if (L'\0' == macro[0])
		return;
	if (NULL == (md = mdmacro_find(mp, macro))) {
		mlerrx(l, MERROR_MACRO, start, L"%ls", macro);
		return;
	}

	/* Bless macro with a type... */

	l->macro = md->type;

	while (L' ' == cp[i])
		i++;
	mline_growarg(l, macro, start);

	/* Branch to a non-generic argument parser. */

	if (NULL != md->parser) {
		(*md->parser)(l, cp, i);
		return;
	} else if (L'\0' == cp[i])
		return;

	/*
	 * Arguments are separated by whitespace.
	 * Quoted arguments begin with a quote (itself preceded by
	 * whitespace or at the beginning)
	 */
	while (L'\0' != cp[i]) {
		start = i + 1;
		if (L'"' == cp[i]) {
			mline_growarg(l, &cp[++i], start + 1);
			while (L'\0' != cp[i] && L'"' != cp[i])
				i++;
			if (L'\0' == cp[i])
				mlerrx(l, MERROR_QUOTE, start, NULL);
			else
				cp[i++] = L'\0';
			continue;
		} else if (L' ' == cp[i]) {
			while (L' ' == cp[i])
				i++;
			continue;
		} 
		mline_growarg(l, &cp[i], i + 1);
		while (L'\0' != cp[i] && 
				(L' ' != cp[i] || L'\\' == cp[i-1]))
			i++;
		if (L'\0' != cp[i])
			cp[i++] = L'\0';
	}
}

/*
 * Parse macro name (`de', `ig', etc.) lines.
 * These are different from generic macros in two ways:
 *   (1) if an escape appears, bail on the entire line
 *   (2) disregard quote mark entirely
 */
void
mline_de(struct mline *l, wchar_t *cp, size_t i)
{

	if (L'\0' == cp[i]) {
		mlerrx(l, MERROR_ARGSFEW, i, NULL);
		return;
	}

	while (L'\0' != cp[i]) {
		if (L' ' == cp[i]) {
			while (L' ' == cp[i])
				i++;
			continue;
		}
		mline_growarg(l, &cp[i], i + 1);
		while (L'\0' != cp[i] && L' ' != cp[i]) {
			if (L'\\' != cp[i++]) 
				continue;
			cp[i - 1] = L'\0';
			mlerrx(l, MERROR_NAMEDEF, i, 
				L"%ls", l->argv[l->argc - 1]);
			return;
		}
		if (L'\0' != cp[i])
			cp[i++] = L'\0';
	}
}

/*
 * Parse a single character (`cc', `c2', etc.).
 * These takes an opaque character as an argument.
 */
void
mline_cc(struct mline *l, wchar_t *cp, size_t i)
{

	if (L'\0' == cp[i])
		return;
	mline_growarg(l, &cp[i], i + 1);
	if (L'\0' != cp[++i])
		mlerrx(l, MERROR_ARGSMANY, i + 1, NULL);
	cp[i] = L'\0';
}

/*
 * Has no arguments at all.
 */
void
mline_ecr(struct mline *l, wchar_t *cp, size_t i)
{

	if (L'\0' != cp[i])
		mlerrx(l, MERROR_ARGSMANY, i, NULL);
}

/*
 * Parse a symbol definition.
 * This consists of a stringvar (name) and its space-separated value,
 * which extends to the EOLN regardless of what it contains.
 * The value begins at the next non-space character (a leading quotation
 * mark will be thrown away).
 */
void
mline_ds(struct mline *l, wchar_t *cp, size_t i)
{

	if (L'\0' == cp[i]) {
		mlerrx(l, MERROR_ARGSFEW, i, NULL);
		return;
	}

	mline_growarg(l, &cp[i], i + 1);

	/* 
	 * We bail if the name contains a slash.
	 * This differs from `de', where we let interrupting escapes
	 * stop our line processing, but continuing with the definition.
	 */
	while (L'\0' != cp[i] && L' ' != cp[i])
		if (l->file->parse->esc == cp[i++]) {
			mlerrx(l, MERROR_STRINGDEF, i, NULL);
			l->argc = 0;
			return;
		}

	if (L'\0' != cp[i])
		cp[i++] = L'\0';
	while (L' ' == cp[i])
		i++;
	if (L'\0' == cp[i])
		return;
	else if (L'"' == cp[i])
		i++;

	mline_growarg(l, &cp[i], i + 1);
}

/*
 * Parse a conditional (`ie', `if').
 * This is rudimentary at best: instead of trying properly to parse the
 * conditional, I simply skip to the next whitespace.
 */
void
mline_if(struct mline *l, wchar_t *cp, size_t i)
{

	if (L'\0' == cp[i]) {
		mlerrx(l, MERROR_ARGSFEW, i, NULL);
		return;
	}

	mline_growarg(l, &cp[i], i + 1);

	while (L'\0' != cp[i] && L' ' != cp[i])
		i++;
	if (L'\0' != cp[i])
		cp[i++] = L'\0';

	while (L' ' == cp[i])
		i++;

	if (L'\0' != cp[i])
		mline_growarg(l, &cp[i], i + 1);
	else
		mlerrx(l, MERROR_ARGSFEW, i, NULL);
}

/*
 * An "else" clause is simply a line containing anything.
 */
void
mline_el(struct mline *l, wchar_t *cp, size_t i)
{

	if (L'\0' == cp[i])
		mlerrx(l, MERROR_ARGSFEW, i, NULL);
	else
		mline_growarg(l, &cp[i], i + 1);
}

/*
 * Create a dynamic line from a string.
 * A dynamic line doesn't point into a file's data buffer.
 * Instead, the "line" variable is NULL and the copy buffer is used
 * instead.
 */
struct mline *
mline_dynalloc(struct mline *l, const wchar_t *cp)
{
	size_t		 sz;
	struct mline	*ml;
	struct mparse	*mp = l->file->parse;

	sz = wcslen(cp);
	ml = mandoc_calloc(1, sizeof(struct mline));
	ml->file = l->file;
	ml->lineno = l->lineno;
	ml->copy = mandoc_malloc((sz + 1) * sizeof(wchar_t));
	wcslcpy(ml->copy, cp, sz + 1);
	ml->dynnext = mp->dynlines;
	mp->dynlines = ml;
	return(ml);
}
