/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/param.h>

#include <assert.h>
#include <stdlib.h>
#include <wchar.h>

#include "mandoc.h"
#include "extern.h"

static	void	 mnode_alloc_root(struct mparse *mp);

static	const enum melem elems[MDTYPE__MAX] = {
	MELEM__TEXT, /* MDTYPE__TEXT */
	MELEM_Ac, /* MDTYPE_Ac */
	MELEM__MAX, /* MDTYPE_Ad */
	MELEM__MAX, /* MDTYPE_An */
	MELEM_Ao, /* MDTYPE_Ao */
	MELEM__MAX, /* MDTYPE_Ap */
	MELEM_Ao, /* MDTYPE_Aq */
	MELEM__MAX, /* MDTYPE_Ar */
	MELEM__MAX, /* MDTYPE_At */
	MELEM__MAX, /* MDTYPE_Bc */
	MELEM__MAX, /* MDTYPE_Bd */
	MELEM__MAX, /* MDTYPE_Bf */
	MELEM__MAX, /* MDTYPE_Bk */
	MELEM__MAX, /* MDTYPE_Bl */
	MELEM__MAX, /* MDTYPE_Bo */
	MELEM__MAX, /* MDTYPE_Bq */
	MELEM__MAX, /* MDTYPE_Brc */
	MELEM__MAX, /* MDTYPE_Bro */
	MELEM__MAX, /* MDTYPE_Brq */
	MELEM__MAX, /* MDTYPE_Bsx */
	MELEM__MAX, /* MDTYPE_Bt */
	MELEM__MAX, /* MDTYPE_Bx */
	MELEM__MAX, /* MDTYPE_Cd */
	MELEM__MAX, /* MDTYPE_Cm */
	MELEM__MAX, /* MDTYPE_D1 */
	MELEM__MAX, /* MDTYPE_Db */
	MELEM__MAX, /* MDTYPE_Dc */
	MELEM__MAX, /* MDTYPE_Dd */
	MELEM__MAX, /* MDTYPE_Dl */
	MELEM__MAX, /* MDTYPE_Do */
	MELEM__MAX, /* MDTYPE_Dq */
	MELEM_Dt, /* MDTYPE_Dt */
	MELEM__MAX, /* MDTYPE_Dv */
	MELEM__MAX, /* MDTYPE_Dx */
	MELEM__MAX, /* MDTYPE_Ec */
	MELEM__MAX, /* MDTYPE_Ed */
	MELEM__MAX, /* MDTYPE_Ef */
	MELEM__MAX, /* MDTYPE_Ek */
	MELEM__MAX, /* MDTYPE_El */
	MELEM__MAX, /* MDTYPE_Em */
	MELEM__MAX, /* MDTYPE_En */
	MELEM__MAX, /* MDTYPE_Eo */
	MELEM__MAX, /* MDTYPE_Er */
	MELEM__MAX, /* MDTYPE_Es */
	MELEM__MAX, /* MDTYPE_Ev */
	MELEM__MAX, /* MDTYPE_Ex */
	MELEM__MAX, /* MDTYPE_Fa */
	MELEM__MAX, /* MDTYPE_Fc */
	MELEM__MAX, /* MDTYPE_Fd */
	MELEM__MAX, /* MDTYPE_Fl */
	MELEM__MAX, /* MDTYPE_Fn */
	MELEM__MAX, /* MDTYPE_Fo */
	MELEM__MAX, /* MDTYPE_Fr */
	MELEM__MAX, /* MDTYPE_Ft */
	MELEM__MAX, /* MDTYPE_Fx */
	MELEM__MAX, /* MDTYPE_Hf */
	MELEM__MAX, /* MDTYPE_Ic */
	MELEM__MAX, /* MDTYPE_In */
	MELEM__MAX, /* MDTYPE_It */
	MELEM__MAX, /* MDTYPE_Lb */
	MELEM__MAX, /* MDTYPE_Li */
	MELEM__MAX, /* MDTYPE_Lk */
	MELEM__MAX, /* MDTYPE_Lp */
	MELEM__MAX, /* MDTYPE_Ms */
	MELEM__MAX, /* MDTYPE_Mt */
	MELEM__MAX, /* MDTYPE_Nd */
	MELEM__MAX, /* MDTYPE_Nm */
	MELEM__MAX, /* MDTYPE_No */
	MELEM__MAX, /* MDTYPE_Ns */
	MELEM__MAX, /* MDTYPE_Nx */
	MELEM_Oc, /* MDTYPE_Oc */
	MELEM_Oo, /* MDTYPE_Oo */
	MELEM_Op, /* MDTYPE_Op */
	MELEM__MAX, /* MDTYPE_Os */
	MELEM__MAX, /* MDTYPE_Ot */
	MELEM__MAX, /* MDTYPE_Ox */
	MELEM__MAX, /* MDTYPE_Pa */
	MELEM__MAX, /* MDTYPE_Pc */
	MELEM__MAX, /* MDTYPE_Pf */
	MELEM__MAX, /* MDTYPE_Po */
	MELEM__MAX, /* MDTYPE_Pp */
	MELEM__MAX, /* MDTYPE_Pq */
	MELEM__MAX, /* MDTYPE_Qc */
	MELEM__MAX, /* MDTYPE_Ql */
	MELEM__MAX, /* MDTYPE_Qo */
	MELEM__MAX, /* MDTYPE_Qq */
	MELEM__MAX, /* MDTYPE_Re */
	MELEM__MAX, /* MDTYPE_Rs */
	MELEM__MAX, /* MDTYPE_Rv */
	MELEM__MAX, /* MDTYPE_Sc */
	MELEM_Sh, /* MDTYPE_Sh */
	MELEM__MAX, /* MDTYPE_Sm */
	MELEM__MAX, /* MDTYPE_So */
	MELEM__MAX, /* MDTYPE_Sq */
	MELEM_Ss, /* MDTYPE_Ss */
	MELEM__MAX, /* MDTYPE_St */
	MELEM__MAX, /* MDTYPE_Sx */
	MELEM__MAX, /* MDTYPE_Sy */
	MELEM__MAX, /* MDTYPE_TH */
	MELEM__MAX, /* MDTYPE_Ta */
	MELEM__MAX, /* MDTYPE_Tn */
	MELEM__MAX, /* MDTYPE_Ud */
	MELEM__MAX, /* MDTYPE_Ux */
	MELEM__MAX, /* MDTYPE_Va */
	MELEM__MAX, /* MDTYPE_Vt */
	MELEM__MAX, /* MDTYPE_Xc */
	MELEM__MAX, /* MDTYPE_Xo */
	MELEM__MAX, /* MDTYPE_Xr */
	MELEM__MAX, /* MDTYPE__A */
	MELEM__MAX, /* MDTYPE__B */
	MELEM__MAX, /* MDTYPE__C */
	MELEM__MAX, /* MDTYPE__D */
	MELEM__MAX, /* MDTYPE__I */
	MELEM__MAX, /* MDTYPE__J */
	MELEM__MAX, /* MDTYPE__N */
	MELEM__MAX, /* MDTYPE__O */
	MELEM__MAX, /* MDTYPE__P */
	MELEM__MAX, /* MDTYPE__Q */
	MELEM__MAX, /* MDTYPE__R */
	MELEM__MAX, /* MDTYPE__T */
	MELEM__MAX, /* MDTYPE__U */
	MELEM__MAX, /* MDTYPE__V */
	MELEM__MAX, /* MDTYPE_br */
	MELEM__MAX, /* MDTYPE_c2 */
	MELEM__MAX, /* MDTYPE_cc */
	MELEM__MAX, /* MDTYPE_de */
	MELEM__MAX, /* MDTYPE_de1 */
	MELEM__MAX, /* MDTYPE_ds */
	MELEM__MAX, /* MDTYPE_ds1 */
	MELEM__MAX, /* MDTYPE_ec */
	MELEM__MAX, /* MDTYPE_ecr */
	MELEM__MAX, /* MDTYPE_ecs */
	MELEM__MAX, /* MDTYPE_el */
	MELEM__MAX, /* MDTYPE_ie */
	MELEM__MAX, /* MDTYPE_if */
	MELEM__MAX, /* MDTYPE_ig */
	MELEM__MAX, /* MDTYPE_so */
	MELEM__MAX, /* MDTYPE_sp */
	MELEM__MAX, /* MDTYPE_while */
	MELEM__MAX, /* MDTYPE__USERDEF */
	MELEM__MAX, /* MDTYPE__NONE */
};

/*
 * Allocate the root node.
 * This is done on-demand.
 */
static void
mnode_alloc_root(struct mparse *mp)
{
	
	assert(NULL == mp->root);
	assert(NULL == mp->node);
	mp->root = mp->node = 
		mandoc_calloc(1, sizeof(struct mnode));
	mp->root->elem = MELEM__ROOT;
}

void
mnode_alloc_node_line(struct mline *l, enum mdtype hint)
{

	mnode_alloc_node(l, 0, hint);
}

/*
 * Allocate a node from position "pos" of line "l".
 * Let "hint" be its macro type: if unsupplied, we'll look up the macro
 * in the mdmacro table and assign it from there.
 * Then set the macro as the current macro and as a child of the
 * existing current macro (and sibling of prior children).
 */
void
mnode_alloc_node(struct mline *l, size_t pos, enum mdtype hint)
{
	struct mnode	*n;
	struct mparse	*mp = l->file->parse;
	struct mdmacro	*md;

	if (NULL == mp->root)
		mnode_alloc_root(mp);

	if (MDTYPE__MAX == hint) {
		md = mdmacro_find(mp, l->argv[pos]);
		assert(NULL != md);
		hint = md->type;
	}

	assert(MDTYPE__TEXT != hint);
	assert(MDTYPE__MAX != hint);
	assert(MDTYPE__USERDEF != hint);
	assert(MDTYPE__NONE != hint);
	n = mandoc_calloc(1, sizeof(struct mnode));
	n->elem = elems[hint];
	n->line = l->lineno;
	n->pos = l->col[pos];

	assert(NULL != mp->node);
	if (NULL != mp->node->first) {
		mp->node->last->next = n;
		mp->node->last = n;
	} else
		mp->node->first = mp->node->last = n;

	n->parent = mp->node;
	mp->node = n;
}

/*
 * Allocate all text to the current node.
 * If we're a macro line, ignore the first text (ostensibly the macro
 * name).
 */
void
mnode_alloc_text_all(struct mline *l)
{
	size_t		 i;

	i = MDTYPE__TEXT == l->macro ? 0 : 1;

	while (i < l->argc)
		mnode_alloc_text(l, i++);
}

/*
 * Allocate a text argument underneath the current node.
 */
void
mnode_alloc_text(struct mline *l, size_t pos)
{
	struct mparse	*mp = l->file->parse;
	struct mnode	*n;

	if (NULL == mp->root)
		mnode_alloc_root(mp);

	n = mandoc_calloc(1, sizeof(struct mnode));
	n->elem = MELEM__TEXT;
	n->line = l->lineno;
	n->pos = l->col[pos];

	if (NULL != mp->node->first) {
		mp->node->last->next = n;
		mp->node->last = n;
	} else
		mp->node->first = mp->node->last = n;

	n->parent = mp->node;
	n->string = mandoc_wcsdup(l->argv[pos]);
}

void
mnode_free_all(struct mnode *n)
{
	struct mnode	*np;

	for ( ; NULL != n; n = np) {
		if (NULL != n->first)
			mnode_free_all(n->first);
		np = n->next;
		free(n->string);
		free(n);
	}
}
