/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>

#include "mandoc.h"
#include "extern.h"

/*
 * Clear up after a parse sequence.
 * MUST be preceded by a call to mparse_init().
 * After this, all internal memory has been freed and the mparse
 * struct may be simply discarded.
 */
void
mparse_clear(struct mparse *mp)
{
	struct mfile	*f;
	struct mline	*l;

	while (NULL != mp->file) {
		f = mp->file->next;
		mfile_free(mp->file);
		mp->file = f;
	}

	while (NULL != mp->dynlines) {
		l = mp->dynlines->dynnext;
		mline_free(mp->dynlines);
		free(mp->dynlines);
		mp->dynlines = l;
	}

	if (mp->root)
		mnode_free_all(mp->root);

	mdmacro_clear(mp);
}

/*
 * Initialise a parse sequence: load the predefined macros, set control
 * characters, and so on.
 * This must be mirrored by mparse_clear().
 */
void
mparse_init(struct mparse *mp, enum mlevel l, mmsg msg)
{

	mp->esc = mp->escs = L'\\';
	mp->ctr = L'.';
	mp->lctr = L'\'';
	mp->level = MLEVEL_OK;
	mp->wlevel = l;
	mp->file = NULL;
	mp->msg = msg;
	mdmacro_init(mp);
	mp->dynlines = NULL;
	mp->cond = 0;
	mp->mode = MMODE_NROFF;
	mp->root = mp->node = NULL;
}

struct mparse *
mparse_alloc(void)
{

	return(mandoc_calloc(1, sizeof(struct mparse)));
}

void
mparse_free(struct mparse *mp)
{

	free(mp);
}

struct mnode *
mparse_result(struct mparse *mp)
{

	return(mp->root);
}

enum mlevel
mparse_parse(struct mparse *mp, const char *f)
{
	struct mline	*l;

	if ( ! mfile_next(mp, f, NULL))
		return(mp->level);

	assert(NULL != mp->file);
	l = (0 == mp->file->sz) ? NULL : &mp->file->lines[0];

	while (NULL != l) 
		mmacro_next(&l);

	return(mp->level);
}
