/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>

#include "mandoc.h"
#include "extern.h"

#define	EXPAND_LIMIT	100 /* how many escapes in one line */

/*
 * Take a string and unescape its interpreted components into "dst".
 * If there are none to unescape, leave "dst" alone.
 * Return the number of escapes made, which may be zero.
 * Start from the copy-buffer if it exists (i.e., we've already been
 * invoked in a `de'), else use the real line.
 * Note: this might adjust l->next if we merge escaped newlines.
 */
size_t
munescape(struct mline *l, wchar_t **dst)
{
	const wchar_t	*cp, *src;
	struct mparse	*mp = l->file->parse;
	struct mline	*ml;
	struct mdmacro	*md;
	size_t		 i, maxl, ssz, nsz, count, 
			 pos, epos, spos, ressz;
	const wchar_t	*res;

	src = (NULL == l->copy) ? l->line : l->copy;
	ssz = wcslen(src);
	pos = count = 0;

	while (NULL != (cp = wcschr(&src[pos], mp->esc))) {
		/*
		 * Always use positional offsets, not pointers.
		 */
		pos = cp - src + 1;

		/*
		 * Account for all characters that we interpret in copy
		 * mode, at this time limited to comments, escaped
		 * escapes, escaped newlines, and special characters.
		 */
		if (mp->esc == src[pos]) {
			/*
			 * An escaped escape.
			 * Here we strip away the outer escape.
			 */
			if (src != *dst) {
				nsz = ssz * sizeof(wchar_t);
				*dst = mandoc_realloc(*dst, nsz);
				wcslcpy(*dst, src, pos);
				wcslcat(*dst, &src[pos], nsz);
				src = *dst;
			} else
				memmove((char *)cp, (char *)&cp[1],
					sizeof(wchar_t) * 
					(ssz - pos + 1));

			ssz--;
			count++;
			continue;
		} else if (L'"' == src[pos]) {
			/*
			 * Strip off comments.
			 * Once we strip away the comment, we no longer
			 * process as the string is subsequently empty.
			 */
			if (src != *dst) {
				nsz = pos * sizeof(wchar_t);
				*dst = mandoc_realloc(*dst, nsz);
				wcslcpy(*dst, src, pos);
			} else 
				(*dst)[pos - 1] = L'\0';

			return(count + 1);
		} else if (L'\0' == src[pos]) {
			/*
			 * Concatenate lines.
			 * This takes the next line (the next virtual
			 * line) and pastes it into our line.
			 * It then readjusts our next line (as defined
			 * by the current stacking) into the next one's.
			 */
			if (NULL == (ml = l->next[l->stackpos])) {
				mlerrx(l, MERROR_EEOF, ssz - 1, NULL);
				return(count);
			}
			cp = NULL == ml->copy ? ml->line : ml->copy;
			epos = wcslen(cp);
			nsz = (ssz + epos) * sizeof(wchar_t);
			if (*dst != src) {
				*dst = mandoc_realloc(*dst, nsz);
				wcslcpy(*dst, src, ssz);
				wcslcat(*dst, cp, nsz);
			} else {
				*dst = mandoc_realloc(*dst, nsz);
				(*dst)[ssz - 1] = L'\0';
				wcslcat(*dst, cp, nsz);
			}
			count++;
			pos = --ssz;
			ssz += epos;
			l->next[l->stackpos] = ml->next[ml->stackpos];
			src = *dst;
			continue;
		} else if (L'*' != src[pos])
			continue;

		/*
		 * We're at a user-defined string.
		 * The third character decides the length of the name of
		 * the string (one, two, or arbitrary-character).
		 * After looking up the string, we'll paste it into the
		 * current string.
		 */
		spos = pos + 1;
		maxl = 1;

		if (L'\0' == src[spos]) {
			mlerrx(l, MERROR_BADESCAPE, pos - 1, NULL);
			return(count);
		} else if (L'(' == src[spos]) {
			spos++;
			maxl = 2;
		} else if (L'[' == src[spos]) {
			spos++;
			maxl = 0;
		}

		epos = spos;
		for (i = 0; 0 == maxl || i < maxl; i++, epos++)
			if (L'\0' == src[epos]) {
				mlerrx(l, MERROR_BADESCAPE, pos, NULL);
				return(count);
			} else if (0 == maxl && L']' == src[epos])
				break;

		/*
		 * Retrieve the replacement string; if it is undefined,
		 * use an empty string.
		 * If it was a `de' ("multiline") string, just use its
		 * first line for now (TODO: append all lines).
		 */
		md = mdmacro_finds(mp, &src[spos], &src[epos]);
		if (NULL == md || NULL == md->first) {
			if (NULL == md)
				mlerrx(l, MERROR_ESCAPE, pos, NULL);
			res = L"";
		} else
			res = NULL == md->first ?  L"" : 
				(NULL == md->first->copy ?  
				 md->first->line : md->first->copy);

		ressz = wcslen(res);
		nsz = sizeof(wchar_t) * (ssz + ressz + 1);
		epos += maxl ? 0 : 1;

		/*
		 * Either copy over the old string to the new one,
		 * replacing the macro spot, or shuffle bytes around in
		 * the existing one.
		 */
		if (*dst != src) {
			*dst = mandoc_realloc(*dst, nsz);
			wcslcpy(*dst, src, pos);
			wcslcat(*dst, res, nsz);
			wcslcat(*dst, &src[epos], nsz);
		} else {
			*dst = mandoc_realloc(*dst, nsz);
			memmove(&(*dst)[pos - 1 + ressz], 
				&(*dst)[epos],
				sizeof(wchar_t) * 
				(ssz - (epos + - 1)));
			memcpy(&(*dst)[pos - 1], 
				res, ressz * sizeof(wchar_t));
		}

		if (EXPAND_LIMIT < ++count) {
			mlerrx(l, MERROR_STACK, pos, NULL);
			break;
		}

		pos--;
		src = *dst;
		ssz = wcslen(src);
	}

	return(count);
}
