/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/param.h>

#include <assert.h>
#include <stdlib.h>
#include <wchar.h>
#include <wctype.h>

#include "mandoc.h"
#include "extern.h"

static	void	 mmacro_close(struct mline **lp, struct mline **last,
			const wchar_t *end);
static	void	 mmacro_cond(struct mline **lp, int *condp);
static	void	 mmacro_cond_anything(struct mline **lp, 
			int cond, size_t pos);
static	void	 mmacro_setchar(struct mline **lp, wchar_t *c, wchar_t v);
static	void	 mmacro_userdef(struct mline **lp);

/*
 * Given a roff block that ends with dot-dot or a custom name (e.g.,
 * `de', `ig'), scan ahead to its close, "end".
 * Set "lp" to the terminating line (e.g., `..') and "last" to the last
 * valid macro line.
 */
static void
mmacro_close(struct mline **lp, 
		struct mline **last, const wchar_t *end)
{
	const wchar_t	*line;
	size_t		 sz = wcslen(end);

	for (*last = *lp, *lp = (*lp)->next[0]; NULL != *lp; 
			*last = *lp, *lp = (*lp)->next[0]) {
		/*
		 * We must unescape the line, as the terminating line
		 * itself may contain a user-defined string.
		 */
		munescape(*lp, &(*lp)->copy);
		line = NULL == (*lp)->copy ? 
			(*lp)->line : (*lp)->copy;
		/*
		 * We start with a period irrespective of how `cc' or
		 * `c2' is set, then have arbitrary whitespace.
		 */
		if (L'.' != *line++) 
			continue;
		while (iswblank(*line))
			line++;
		/*
		 * The terminating name may have arbitrary trailing
		 * whitespace to the word.
		 */
		if (0 == wcsncmp(line, end, sz)) {
			line += sz;
			while (iswblank(*line))
				line++;
			if (L'\0' == *line)
				break;
		}
	}
}

/*
 * Evaluate the scope block of a conditional.
 * "Scope" blocks are those within `\{' and `\}'.
 * Note that troff(1) is pretty stupid, and thus so are we: within a
 * conditional we must skip over, `\{' anywhere on the line increments
 * the scope, while `\}' decrements it.
 * The scope terms cannot be invoked with `ds' or `de' variables.
 */
static void
mmacro_cond_anything(struct mline **lp, int cond, size_t pos)
{
	struct mline	*l = *lp, *ml;
	struct mparse	*mp = (*lp)->file->parse;
	size_t	  	 scope;
	const wchar_t	*cp, *cpp;
	wchar_t	 	 open[3], close[3];

	/* Use the escape character for the open/close token. */

	open[0] = close[0] = mp->esc;
	open[2] = close[2] = L'\0';
	open[1] = L'{';
	close[1] = L'}';

	/* 
	 * Look for the multiline control characters.
	 * If not found, paste and continue (or just continue).
	 */
	cp = NULL == l->copy ? l->line : l->copy;
	assert(NULL != cp);

	if (NULL == wcsstr(cp, open)) {
		*lp = mline_popnext(*lp);
		if (0 == cond)
			return;
		assert(NULL != l->wline);
		ml = mline_dynalloc(l, &l->wline[pos - 1]);
		ml->next[0] = *lp;
		*lp = ml;
		return;
	} 

	/*
	 * If we're multiline and the conditional is true, then jump
	 * directly into the pasted line.
	 * We don't care about scope, in this case.
	 * Skip over the leading \{ stuff, if found. 
	 */
	if (cond) {
		assert(NULL != l->wline);
		if (0 == wcsncmp(&l->wline[pos], open, 2))
			pos += 2;

		ml = mline_dynalloc(l, &l->wline[pos - 1]);
		ml->next[0] = mline_popnext(*lp);
		*lp = ml;
		return;
	}

	/*
	 * If the condition evaluates to true, we just blindly go
	 * forward, stripping out scopes.  
	 * In each line, we maintain the number of open scopes.
	 * Stop when it equals zero (make sure we don't underrun the
	 * unsigned integer).
	 * Note (0 == scope) is a postconditional because we begin
	 * without scope, then process the current line.
	 */
	for (scope = 0; NULL != *lp; *lp = mline_popnext(*lp)) {
		cp = NULL == (*lp)->copy ?  (*lp)->line : (*lp)->copy;
		assert(NULL != cp);

		cpp = cp;
		for ( ; NULL != (cpp = wcsstr(cpp, open)); cpp += 2)
			scope++;

		cpp = cp;
		for ( ; NULL != (cpp = wcsstr(cpp, close)); cpp += 2)
			scope = scope > 0 ? scope - 1 : 0;

		if (0 == scope)
			break;
	}

	if (NULL != *lp)
		*lp = mline_popnext(*lp);
	else
		mlerrx(l, MERROR_EXITSCOPE, l->col[0], NULL);
}

/*
 * Process a conditional statement.
 * If the condition evaluates to true, continue as if nothing happened
 * (except for pasting the next line string).
 * If, however, it's false, then read til the end of the scope.
 * Use mmacro_cond_anything() for scope rules.
 */
static void
mmacro_cond(struct mline **lp, int *condp)
{
	struct mparse	*mp = (*lp)->file->parse;
	struct mline	*l = *lp;
	const wchar_t	*cp;
	int		 cond = 0, neg;

	if (l->argc < 3) {
		*lp = mline_popnext(*lp);
		return;
	} 

	/*
	 * Evaluate conditional.
	 * For now, accept only "n", "t", and negation "!".
	 */
	cp = l->argv[1];
	if (0 != (neg = L'!' == *cp))
		cp++;
	if (0 == wcscmp(cp, L"n"))
		cond = MMODE_NROFF == mp->mode;
	else if (0 == wcscmp(cp, L"t"))
		cond = MMODE_TROFF == mp->mode;
	if (neg)
		cond = !cond;
	if (NULL != condp)
		*condp = cond;

	mmacro_cond_anything(lp, cond, l->col[2]);
}

/*
 * A user-defined macro is one that's provided with `ds' or `de'.
 * (Note that both macros and strings have the same underlying machinery
 * according to groff.texinfo.)
 */
static void
mmacro_userdef(struct mline **lp)
{
	struct mparse	*mp = (*lp)->file->parse;
	struct mdmacro	*md;

	md = mdmacro_find(mp, (*lp)->argv[0]);
	assert(NULL != md);

	if (NULL != md->first) {
		if (mline_pushnext(md->last, mline_peek(*lp))) {
			mline_popnext(*lp);
			*lp = md->first;
		} else
			*lp = mline_popnext(*lp);
	} else
		*lp = mline_popnext(*lp);
}

/*
 * Convenience function for setting the character "c" from the first
 * argument of macro or "v" if a first character doesn't exist.
 */
static void
mmacro_setchar(struct mline **lp, wchar_t *c, wchar_t v)
{

	*c = 1 == (*lp)->argc ? v : (*lp)->argv[1][0];
	*lp = mline_popnext(*lp);
}

void
mmacro_char(struct mline **lp)
{

	switch ((*lp)->macro) {
	case (MDTYPE_cc):
		mmacro_setchar(lp, &(*lp)->file->parse->ctr, L'.');
		break;
	case (MDTYPE_c2):
		mmacro_setchar(lp, &(*lp)->file->parse->lctr, L'\'');
		break;
	case (MDTYPE_ec):
		mmacro_setchar(lp, &(*lp)->file->parse->esc, L'\\');
		break;
	case (MDTYPE_ecs):
		mmacro_setchar(lp, &(*lp)->file->parse->escs, L'\\');
		break;
	default:
		abort();
	}
}

void
mmacro_ecr(struct mline **lp)
{

	(*lp)->file->parse->esc = (*lp)->file->parse->escs;
	*lp = mline_popnext(*lp);
}

void
mmacro_el(struct mline **lp)
{
	struct mparse	*mp = (*lp)->file->parse;
	struct mline	*l = *lp;

	if (l->argc < 2)
		*lp = mline_popnext(*lp);
	else
		mmacro_cond_anything(lp, !mp->cond, l->col[1]);
}

void
mmacro_while(struct mline **lp)
{
	struct mline	*l = *lp;

	/*
	 * Ignore by assuming "false" and routing directly into the
	 * "anything" handler of the conditional.
	 */
	if (l->argc < 3)
		*lp = mline_popnext(*lp);
	else
		mmacro_cond_anything(lp, 0, l->col[2]);
}

void
mmacro_if(struct mline **lp)
{

	mmacro_cond(lp, NULL);
}

void
mmacro_ie(struct mline **lp)
{

	mmacro_cond(lp, &(*lp)->file->parse->cond);
}

void
mmacro_ig(struct mline **lp)
{
	struct mline	*first = *lp, *last;

	mmacro_close(lp, &last, (*lp)->argc > 1 && 
		L'\0' != (*lp)->argv[1] ? (*lp)->argv[1] : L".");

	if (NULL != *lp)
		*lp = mline_popnext(*lp);
	else
		mlerrx(first, MERROR_EXITSCOPE, first->col[0], NULL);
}

void
mmacro_ds(struct mline **lp)
{
	struct mline	*l = *lp, *ml;
	struct mparse	*mp = (*lp)->file->parse;
	struct mdmacro	*md;
	unsigned int	 index;

	*lp = mline_popnext(l);

	if (l->argc < 2)
		return;

	index = mdmacro_lookup(mp, l->argv[1]);
	if (NULL == (md = ohash_find(&mp->macros, index))) {
		md = mdmacro_calloc(l->argv[1]);
		ohash_insert(&mp->macros, index, md);
	} 

	md->macro = mmacro_userdef;
	md->first = md->last = NULL;
	md->parser = NULL;

	if (l->argc < 3)
		return;

	assert(NULL != l->wline);
	ml = mline_dynalloc(l, l->wline + l->col[2] - 1);
	md->first = md->last = ml;
}

void
mmacro_de(struct mline **lp)
{
	struct mparse	*mp = (*lp)->file->parse;
	struct mline	*first = *lp, *last;
	struct mdmacro	*md;
	unsigned int	 index;

	mmacro_close(lp, &last, (*lp)->argc > 2 && 
		L'\0' != (*lp)->argv[2] ? (*lp)->argv[2] : L".");

	if (NULL == *lp) {
		mlerrx(first, MERROR_EXITSCOPE, first->col[0], NULL);
		return;
	}

	*lp = mline_popnext(*lp);
	index = mdmacro_lookup(mp, first->argv[1]);
	if (NULL == (md = ohash_find(&mp->macros, index))) {
		md = mdmacro_calloc(first->argv[1]);
		ohash_insert(&mp->macros, index, md);
	} 

	md->macro = mmacro_userdef;
	md->type = MDTYPE__USERDEF;
	md->first = md->last = NULL;
	md->parser = NULL;

	if (first != last) {
		md->first = first->next[0];
		md->last = last;
	} 
}

void
mmacro_so(struct mline **lp)
{
	struct mline	*l = *lp;
	char		 file[MAXPATHLEN];
	struct mparse	*mp = (*lp)->file->parse;
	size_t		 sz;

	if (l->argc < 2) {
		*lp = mline_popnext(l);
		return;
	} 

	/* Convert the wchar_t filename back into char. */

	*lp = NULL;
	sz = wcstombs(file, l->argv[1], MAXPATHLEN);
	if ((size_t)-1 != sz && sz < MAXPATHLEN) {
		if (mfile_next(l->file->parse, file, l))
			*lp = mline_popnext(l); 
	} else 
		merrx(mp, MERROR_WCS, l->file->fname, 0, 0, NULL);
}
